# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack import proxy
from openstack import resource

from otcextensions.common.utils import extract_url_parts
from otcextensions.sdk.apig.v2 import gateway as _gateway
from otcextensions.sdk.apig.v2 import az as _az
from otcextensions.sdk.apig.v2 import apienvironment as _api_environment
from otcextensions.sdk.apig.v2 import apigroup as _api_group
from otcextensions.sdk.apig.v2 import apienvironmentvar as _api_var
from otcextensions.sdk.apig.v2 import throttling_policy as _tp
from otcextensions.sdk.apig.v2 import api as _api
from otcextensions.sdk.apig.v2 import api_supplements as _supp
from otcextensions.sdk.apig.v2 import app as _app
from otcextensions.sdk.apig.v2 import appcode as _app_code
from otcextensions.sdk.apig.v2 import signature as _sign
from otcextensions.sdk.apig.v2 import signature_binding as _sign_bind
from otcextensions.sdk.apig.v2 import throttling_policy_binding as _tpb
from otcextensions.sdk.apig.v2 import throttling_excluded as _tx
from otcextensions.sdk.apig.v2 import gateway_features as _gwf
from otcextensions.sdk.apig.v2 import resource_query as _rq
from otcextensions.sdk.apig.v2 import domain_name as _domain
from otcextensions.sdk.apig.v2 import certificate as _c
from otcextensions.sdk.apig.v2 import api_auth as _auth
from otcextensions.sdk.apig.v2 import acl_policy as _acl
from otcextensions.sdk.apig.v2 import acl_api_binding as _acl_api_binding
from otcextensions.sdk.apig.v2 import custom_authorizer as _custom_auth
from otcextensions.sdk.apig.v2 import export_api as _export_api
from otcextensions.sdk.apig.v2 import vpc_channel as _vpc_channel
from otcextensions.sdk.apig.v2 import backend_server_group as _backend_group
from otcextensions.sdk.apig.v2 import backend_server as _backend_server
from otcextensions.sdk.apig.v2 import api_call as _api_call
from otcextensions.sdk.apig.v2 import metric_data as _metric_data


class Proxy(proxy.Proxy):
    skip_discovery = True

    def _extract_name(self, url, service_type=None, project_id=None):
        return extract_url_parts(url, project_id)

    def gateways(self, **attrs):
        """Query gateways

        :returns: A generator of gateway object of a
            :class:`~otcextensions.sdk.apig.gateway.Gateway`
        """
        return self._list(_gateway.Gateway, paginated=False, **attrs)

    def create_gateway(self, **attrs):
        """Create gateway

        :returns: A gateway object
        :rtype: :class:`~otcextensions.sdk.apig.gateway.Gateway`
        """
        return self._create(_gateway.Gateway, **attrs)

    def wait_for_gateway(self, gateway, status='Running', failures=None,
                         interval=2, wait=960):
        """Wait for specific gateway status
        :param gateway: key id or an instance of
        :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :param status: Specific status of the gateway.
        :param failures: Specific failure status of the gateway.
        :param interval: Seconds between checking the gateway.
        :param wait: Seconds between checking the gateway.

        %returns: instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        if failures is None:
            failures = ['ERROR']
        return resource.wait_for_status(
            self, gateway, status, failures, interval, wait)

    def get_gateway(self, gateway):
        """Get details of specific gateway
        :param gateway: key id or an instance of
        :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :returns: instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        """
        return self._get(_gateway.Gateway, gateway)

    def update_gateway(self, gateway, **attrs):
        """Update an existing API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: The attributes to update on the gateway.

        :returns: Updated instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        """
        return self._update(_gateway.Gateway, gateway, **attrs)

    def delete_gateway(self, gateway, **attrs):
        """Delete specific gateway
        :param gateway: key id or an instance of
        :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :returns: 'None'
        """
        return self._delete(_gateway.Gateway, gateway, **attrs)

    def get_gateway_progress(self, gateway):
        """Get specific gateway progress
        :param gateway: key id or an instance of
        :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        %returns: instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return gateway._get_creation_progress(self, gateway)

    def get_constraints(self, gateway):
        """Get gateway constraints

        :param gateway: key id or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        %returns: instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return gateway._get_constraints(self, gateway)

    def enable_public_access(self, gateway, **attrs):
        """Enable public access for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional attributes to configure public access.

        :returns: Updated instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return gateway._enable_public_access(self, gateway, **attrs)

    def update_public_access(self, gateway, **attrs):
        """Update public access settings for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: The attributes to update public access settings.

        :returns: Updated instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return gateway._update_public_access(self, gateway, **attrs)

    def disable_public_access(self, gateway):
        """Disable public access for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :returns: Updated instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return gateway._disable_public_access(self, gateway)

    def modify_gateway_spec(self, gateway, **attrs):
        """Modify the specifications of a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: The attributes to modify the gateway specifications.

        :returns: Updated instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return gateway._modify_spec(self, gateway, **attrs)

    def bind_eip(self, gateway, **attrs):
        """Bind an Elastic IP (EIP) to a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: The attributes required for binding the EIP.

        :returns: Updated instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return gateway._bind_eip(self, gateway, **attrs)

    def unbind_eip(self, gateway):
        """Unbind an Elastic IP (EIP) from a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :returns: Updated instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return gateway._unbind_eip(self, gateway)

    def azs(self, **attrs):
        """Retrieve availability zones (AZs) for API Gateway service.

        :param attrs: Optional filters for querying availability zones.

        :returns: A list of availability zones.
        :rtype: list of :class:`~otcextensions.sdk.apig.v2.az.AZ`
        """
        return self._list(_az.AZ, paginated=False, **attrs)

    def enable_ingress(self, gateway, **attrs):
        """Enable public inbound access for a specific API Gateway.

        This method binds a public IP to an API Gateway to allow
        public inbound traffic from the internet.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :param attrs: Additional attributes required for enabling public
            inbound access

        :returns: Updated instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return gateway._enable_ingress(self, gateway, **attrs)

    def update_ingress(self, gateway, **attrs):
        """Update public inbound access bandwidth of a specific API Gateway.

        This method modifies the inbound bandwidth settings for an API Gateway
        that has public access enabled.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :param attrs: Additional attributes required for updating the ingress
            bandwidth

        :returns: Updated instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return gateway._update_ingress(self, gateway, **attrs)

    def disable_ingress(self, gateway):
        """Disable public inbound access for a specific API Gateway.

        This method removes public inbound access from an API Gateway
        by unbinding the associated public IP.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :returns: Updated instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return gateway._disable_ingress(self, gateway)

    def create_environment(self, gateway, **attrs):
        """Create a new environment for a specific API Gateway.

        This method creates an environment within the given API Gateway
        by associating it with the specified attributes.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional attributes for the environment creation.

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.api_environment.ApiEnvironment`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._create(_api_environment.ApiEnvironment,
                            gateway_id=gateway.id,
                            **attrs)

    def update_environment(self, gateway, environment, **attrs):
        """Update an existing environment for a specific API Gateway.

        This method updates the specified environment within the API Gateway
        by applying the provided attributes.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param environment: The ID of the environment or an instance of
            :class:`~otcextensions.sdk.apig.v2.environment.ApiEnvironment`
        :param attrs: Additional attributes to update the environment.

        :returns: Updated instance of
            :class:`~otcextensions.sdk.apig.v2.environment.ApiEnvironment`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        environment = self._get_resource(_api_environment.ApiEnvironment,
                                         environment)
        return environment._update_env(self, gateway, **attrs)

    def delete_environment(self, gateway, environment, **attrs):
        """Delete an existing environment from a specific API Gateway.

        This method removes the specified environment from the API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param environment: The ID of the environment or an instance of
            :class:`~otcextensions.sdk.apig.v2.environment.ApiEnvironment`
        :param attrs: Additional attributes for the delete operation.

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        environment = self._get_resource(_api_environment.ApiEnvironment,
                                         environment)
        return self._delete(_api_environment.ApiEnvironment,
                            environment,
                            gateway_id=gateway.id,
                            **attrs)

    def environments(self, gateway, **attrs):
        """List all environments for a specific API Gateway.

        This method retrieves a list of environments associated with
        the given API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional filters for listing environments.

        :returns: A list of instances of
            :class:`~otcextensions.sdk.apig.v2.environment.ApiEnvironment`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(_api_environment.ApiEnvironment,
                          paginated=False,
                          gateway_id=gateway.id,
                          **attrs)

    def create_api_group(self, gateway, **attrs):
        """Create a new API group for a specific API Gateway.

        This method creates an API group associated with the given API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional attributes for creating the API group.

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.api_group.ApiGroup`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._create(_api_group.ApiGroup,
                            gateway_id=gateway.id,
                            **attrs)

    def update_api_group(self, gateway, api_group, **attrs):
        """Update an existing API group for a specific API Gateway.

        This method updates the attributes of an API group associated with
        the given API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param api_group: The ID of the API group or an instance of
            :class:`~otcextensions.sdk.apig.v2.api_group.ApiGroup`
        :param attrs: Additional attributes for updating the API group.

        :returns: The updated instance of
            :class:`~otcextensions.sdk.apig.v2.api_group.ApiGroup`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        api_group = self._get_resource(_api_group.ApiGroup, api_group)
        return api_group._update_group(self, gateway=gateway, **attrs)

    def delete_api_group(self, gateway, api_group, **attrs):
        """Delete an API group from a specific API Gateway.

        This method deletes the specified API group associated with
        the given API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param api_group: The ID of the API group or an instance of
            :class:`~otcextensions.sdk.apig.v2.api_group.ApiGroup`
        :param attrs: Additional parameters for deleting the API group.

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        api_group = self._get_resource(_api_group.ApiGroup, api_group)
        return self._delete(_api_group.ApiGroup,
                            api_group,
                            gateway_id=gateway.id,
                            **attrs)

    def get_api_group(self, gateway, api_group):
        """Retrieve details of a specific API group.

        This method retrieves the details of an API group associated
        with the given API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param api_group: The ID of the API group or an instance of
            :class:`~otcextensions.sdk.apig.v2.api_group.ApiGroup`

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.api_group.ApiGroup`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._get(_api_group.ApiGroup,
                         api_group,
                         gateway_id=gateway.id)

    def api_groups(self, gateway, **attrs):
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(_api_group.ApiGroup,
                          paginated=False,
                          gateway_id=gateway.id,
                          **attrs)

    def verify_api_group_name(self, gateway, **attrs):
        gateway = self._get_resource(_gateway.Gateway, gateway)
        api_group = _api_group.ApiGroup()
        return api_group._verify_name(self, gateway=gateway, **attrs)

    # ======== Environment Variable Methods ========

    def create_environment_variable(self, gateway, **attrs):
        """Create a new environment variable for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional attributes for the environment variable
            creation.

        :returns: An instance of ApiEnvironmentVar
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._create(_api_var.ApiEnvironmentVar,
                            gateway_id=gateway.id,
                            **attrs)

    def update_environment_variable(self, gateway, var, **attrs):
        """Update an existing environment variable for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param var: The ID of the environment var or an instance of
            ApiEnvironmentVar
        :param attrs: Additional attributes to update the environment.

        :returns: Updated instance of ApiEnvironmentVar
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        var = self._get_resource(
            _api_var.ApiEnvironmentVar,
            var)
        return self._update(
            _api_var.ApiEnvironmentVar,
            var,
            gateway_id=gateway.id,
            **attrs
        )

    def delete_environment_variable(self, gateway, var, ignore_missing=False):
        """Delete an existing environment variable from a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param var: The ID of the environment or an instance of
            ApiEnvironmentVar

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        var = self._get_resource(
            _api_var.ApiEnvironmentVar,
            var)
        return self._delete(
            _api_var.ApiEnvironmentVar,
            var,
            gateway_id=gateway.id,
            ignore_missing=ignore_missing
        )

    def environment_variables(self, gateway, **attrs):
        """List all environment vars for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional filters for listing environment vars.

        :returns: A list of instances of ApiEnvironmentVar
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(_api_var.ApiEnvironmentVar,
                          paginated=False,
                          gateway_id=gateway.id,
                          **attrs)

    def get_environment_variable(self, gateway, var):
        """Retrieve details of a specific environment variable.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param var: The ID of the variable or an instance of
            ApiEnvironmentVar

        :returns: An instance of ApiEnvironmentVar
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        var = self._get_resource(
            _api_var.ApiEnvironmentVar,
            var
        )
        return self._get(
            _api_var.ApiEnvironmentVar,
            var,
            gateway_id=gateway.id
        )

    # ======== Throttling Policy Methods ========

    def create_throttling_policy(self, gateway, **attrs):
        """Create a new throttling policy for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional attributes for the throttling policy
            creation.

        :returns: An instance of ThrottlingPolicy
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._create(_tp.ThrottlingPolicy,
                            gateway_id=gateway.id,
                            **attrs)

    def update_throttling_policy(self, gateway, policy, **attrs):
        """Update an existing throttling policy for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param tp: The ID of the throttling policy or an instance of
            ThrottlingPolicy
        :param attrs: Additional attributes to update the throttling policy.

        :returns: Updated instance of ThrottlingPolicy
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        policy = self._get_resource(
            _tp.ThrottlingPolicy,
            policy)
        return self._update(
            _tp.ThrottlingPolicy,
            policy,
            gateway_id=gateway.id,
            **attrs
        )

    def delete_throttling_policy(self, gateway, policy, ignore_missing=False):
        """Delete an existing throttling policy from a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param policy: The ID of the throttling policy or an instance of
            ThrottlingPolicy

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        policy = self._get_resource(
            _tp.ThrottlingPolicy,
            policy)
        return self._delete(
            _tp.ThrottlingPolicy,
            policy,
            gateway_id=gateway.id,
            ignore_missing=ignore_missing
        )

    def throttling_policies(self, gateway, **attrs):
        """List all throttling policies for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional filters for listing throttling policies.

        :returns: A list of instances of ThrottlingPolicy
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(_tp.ThrottlingPolicy,
                          paginated=False,
                          gateway_id=gateway.id,
                          **attrs)

    def get_throttling_policy(self, gateway, policy):
        """Retrieve details of a specific throttling policy.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param policy: The ID of the throttling policy or an instance of
            ThrottlingPolicy

        :returns: An instance of ThrottlingPolicy
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        policy = self._get_resource(
            _tp.ThrottlingPolicy,
            policy
        )
        return self._get(
            _tp.ThrottlingPolicy,
            policy,
            gateway_id=gateway.id
        )

    # ======== Api Methods ========

    def create_api(self, gateway, **attrs):
        """Create a new API for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional attributes for the Api creation.

        :returns: An instance of Api
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._create(_api.Api,
                            gateway_id=gateway.id,
                            **attrs)

    def update_api(self, gateway, api, **attrs):
        """Update an existing API for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param api: The ID of the Api or an instance of Api
        :param attrs: Additional attributes to update the Api.

        :returns: Updated instance of Api
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        api = self._get_resource(_api.Api, api)
        return self._update(
            _api.Api,
            api,
            gateway_id=gateway.id,
            **attrs
        )

    def delete_api(self, gateway, api, ignore_missing=False):
        """Delete an existing API from a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param api: The ID of the Api or an instance of Api

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        api = self._get_resource(_api.Api, api)
        return self._delete(
            _api.Api,
            api,
            gateway_id=gateway.id,
            ignore_missing=ignore_missing
        )

    def apis(self, gateway, **attrs):
        """List all APIs for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional filters for listing Api.

        :returns: A list of instances of Api
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(_api.Api,
                          paginated=False,
                          gateway_id=gateway.id,
                          **attrs)

    def get_api(self, gateway, api):
        """Retrieve details of a specific API.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param api: The ID of the Api or an instance of Api

        :returns: An instance of Api
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        api = self._get_resource(_api.Api, api)
        return self._get(
            _api.Api,
            api,
            gateway_id=gateway.id
        )

    def publish_api(self, gateway, env, api, **attrs):
        """Publish an API.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param api: The ID of the Api or an instance of Api
        :param env: The ID of the Environment or an instance of it
        :param attrs: Additional attributes

        :returns: An instance of PublishApi
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        env = self._get_resource(_api_environment.ApiEnvironment, env)
        api = self._get_resource(_api.Api, api)
        action = self._get_resource(_supp.PublishApi, "")
        return action.publish_api(
            self,
            api_id=api.id,
            env_id=env.id,
            gateway_id=gateway.id,
            **attrs
        )

    def offline_api(self, gateway, env, api, **attrs):
        """Take API offline.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param api: The ID of the Api or an instance of Api
        :param env: The ID of the Environment or an instance of it
        :param attrs: Additional attributes

        :returns: An instance of PublishApi
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        env = self._get_resource(_api_environment.ApiEnvironment, env)
        api = self._get_resource(_api.Api, api)
        action = self._get_resource(_supp.PublishApi, "")
        return action.take_api_offline(
            self,
            api_id=api.id,
            env_id=env.id,
            gateway_id=gateway.id,
            **attrs
        )

    def check_api(self, gateway, **attrs):
        """Verify the API definition.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional attributes

        :returns: An instance of CheckApi
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._create(
            _supp.CheckApi,
            gateway_id=gateway.id,
            **attrs)

    def debug_api(self, gateway, api, **attrs):
        """Debug an API in a specified environment.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param api: The ID of the Api or an instance of Api
        :param attrs: Additional attributes

        :returns: An instance of DebugApi
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        api = self._get_resource(_api.Api, api)
        return self._create(
            _supp.DebugApi,
            gateway_id=gateway.id,
            api_id=api.id,
            **attrs)

    def publish_apis(self, gateway, env, **attrs):
        """Publish an APIs.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param env: The ID of the Environment or an instance of it
        :param attrs: Additional attributes

        :returns: An instance of PublishApis
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        env = self._get_resource(_api_environment.ApiEnvironment, env)
        action = self._get_resource(_supp.PublishApis, "")
        return action.publish_apis(
            self,
            env_id=env.id,
            gateway_id=gateway.id,
            **attrs
        )

    def offline_apis(self, gateway, env, **attrs):
        """Takes offline an APIs.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param env: The ID of the Environment or an instance of it
        :param attrs: Additional attributes

        :returns: An instance of PublishApis
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        env = self._get_resource(_api_environment.ApiEnvironment, env)
        action = self._get_resource(_supp.PublishApis, "")
        return action.take_apis_offline(
            self,
            env_id=env.id,
            gateway_id=gateway.id,
            **attrs
        )

    def api_versions(self, gateway, api):
        """Retrieve the historical versions of an API.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param api: The ID of the Api or an instance of Api

        :returns: An instance of PublishApis
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        api = self._get_resource(_supp.PublishApis, api)
        base_path = f'/apigw/instances/%(gateway_id)s/apis/publish/%(api_id)s'
        return self._list(
            _supp.PublishApis,
            api_id=api.id,
            gateway_id=gateway.id,
            base_path=base_path,
        )

    def switch_version(self, gateway, api, version_id):
        """Switch the version of an API.

        :param version_id: API version ID.
        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param api: The ID of the Api or an instance of Api

        :returns: An instance of PublishApis
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        api = self._get_resource(_supp.PublishApis, api)
        return self._update(
            _supp.PublishApis,
            id=api.id,
            gateway_id=gateway.id,
            version_id=version_id,
        )

    def api_runtime_definitions(self, gateway, api, **query):
        """Retrieve the runtime definition of an API.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param api: The ID of the Api or an instance of Api

        :returns: An instance of RuntimeDefinitionApi
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        api = self._get_resource(_supp.PublishApis, api)
        return self._list(
            _supp.RuntimeDefinitionApi,
            api_id=api.id,
            gateway_id=gateway.id,
            **query
        )

    def api_version_details(self, gateway, version_id):
        """Retrieve the details of specified API version.

        :param version_id: API version.
        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :returns: An instance of RuntimeDefinitionApi
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _supp.VersionsApi,
            gateway_id=gateway.id,
            version_id=version_id
        )

    def take_api_version_offline(
            self, gateway, version_id, ignore_missing=False
    ):
        """Remove an effective version of an API.

        :param version_id: API version.
        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._delete(
            _supp.VersionsApi,
            gateway_id=gateway.id,
            version_id=version_id,
            ignore_missing=ignore_missing
        )

    # ======== Signature Keys Methods ========

    def create_signature(self, gateway, **attrs):
        """Create a new Signature for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional attributes for the Signature creation.

        :returns: An instance of Signature
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._create(_sign.Signature,
                            gateway_id=gateway.id,
                            **attrs)

    def update_signature(self, gateway, sign, **attrs):
        """Update an existing Signature for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param sign: The ID of the Signature or an instance of Signature
        :param attrs: Additional attributes to update the Signature.

        :returns: Updated instance of Signature
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        sign = self._get_resource(_sign.Signature, sign)
        return self._update(
            _sign.Signature,
            sign,
            gateway_id=gateway.id,
            **attrs
        )

    def delete_signature(self, gateway, sign, ignore_missing=False):
        """Delete an existing Signature from a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param sign: The ID of the Signature or an instance of Signature

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        sign = self._get_resource(_sign.Signature, sign)
        return self._delete(
            _sign.Signature,
            sign,
            gateway_id=gateway.id,
            ignore_missing=ignore_missing
        )

    def signatures(self, gateway, **attrs):
        """List all Signatures for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional filters for listing Signature.

        :returns: A list of instances of Signature
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _sign.Signature,
            paginated=False,
            gateway_id=gateway.id,
            **attrs
        )

    # ======== Signature Binding Methods ========

    def bind_signature(self, gateway, **attrs):
        """Bind a Signature for a specific API.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional attributes for the Signature bind.

        :returns: An instance of SignatureBind
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._create(
            _sign_bind.SignatureBind,
            gateway_id=gateway.id,
            **attrs)

    def unbind_signature(self, gateway, bind, ignore_missing=False):
        """Unbind a bound Signature from a specific API.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param bind: The ID of the SignatureBind or an instance
            of SignatureBind

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        bind = self._get_resource(_sign_bind.SignatureBind, bind)
        return self._delete(
            _sign_bind.SignatureBind,
            bind,
            gateway_id=gateway.id,
            ignore_missing=ignore_missing
        )

    def bound_signatures(self, gateway, **query):
        """List all Signatures bound a specific API.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param query: Additional filters for listing SignatureBind.

        :returns: A list of instances of SignatureBind
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        bp = '/apigw/instances/%(gateway_id)s/sign-bindings/binded-signs'
        return self._list(
            _sign_bind.SignatureBind,
            paginated=False,
            gateway_id=gateway.id,
            base_path=bp,
            **query
        )

    def not_bound_apis(self, gateway, **query):
        """List all APIs to which a signature key has not been bound.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param query: Additional filters for listing NotBoundApi.

        :returns: A list of instances of NotBoundApi
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _sign_bind.NotBoundApi,
            paginated=False,
            gateway_id=gateway.id,
            **query
        )

    def bound_apis(self, gateway, **query):
        """List all APIs to which a signature key has been bound.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param query: Additional filters for listing BoundApi.

        :returns: A list of instances of BoundApi
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _sign_bind.BoundApi,
            paginated=False,
            gateway_id=gateway.id,
            **query
        )

    # ======== Throttling Policy Binding Methods ========

    def bind_throttling_policy(self, gateway, **attrs):
        """Bind a throttling policy to a specific API.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional attributes for the ThrottlingPolicy bind.

        :returns: An instance of ThrottlingPolicyBind
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._create(
            _tpb.ThrottlingPolicyBind,
            gateway_id=gateway.id,
            **attrs)

    def unbind_throttling_policy(self, gateway, bind, ignore_missing=False):
        """Unbind a bound Signature from a specific API.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param bind: The ID of the SignatureBind or an instance
            of ThrottlingPolicyBind

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        bind = self._get_resource(_tpb.ThrottlingPolicyBind, bind)
        return self._delete(
            _tpb.ThrottlingPolicyBind,
            bind,
            gateway_id=gateway.id,
            ignore_missing=ignore_missing
        )

    def unbind_throttling_policies(self, gateway, throttle_bindings: list):
        """Unbind a bound Signature from a specific API.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param throttle_bindings: The IDs of the request throttling
            policy binding records to be canceled.

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        bind = self._get_resource(_tpb.ThrottlingPolicyBind, "")
        return bind.unbind_policies(
            self,
            gateway_id=gateway.id,
            throttle_bindings=throttle_bindings
        )

    def bound_throttling_policy_apis(self, gateway, **query):
        """List all APIs to which a specified request
            throttling policy has been bound.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param query: Additional filters for listing ThrottlingPolicyBind.

        :returns: A list of instances of ThrottlingPolicyBind
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        bp = '/apigw/instances/%(gateway_id)s/throttle-bindings/binded-apis'
        return self._list(
            _tpb.ThrottlingPolicyBind,
            paginated=False,
            gateway_id=gateway.id,
            base_path=bp,
            **query
        )

    def not_bound_throttling_policy_apis(self, gateway, **query):
        """List all APIs to which a request throttling
            policy has not been bound.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param query: Additional filters for listing NotBoundApi.

        :returns: A list of instances of NotBoundApi
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _tpb.NotBoundApi,
            paginated=False,
            gateway_id=gateway.id,
            **query
        )

    def bound_throttling_policies(self, gateway, **query):
        """List all throttling policies that have been bound to an API.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param query: Additional filters for listing BoundApi.

        :returns: A list of instances of BoundThrottles
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _tpb.BoundThrottles,
            paginated=False,
            gateway_id=gateway.id,
            **query
        )

    # ======== Throttling Policy Methods ========

    def create_throttling_excluded_policy(self, gateway, policy, **attrs):
        """Creating an Excluded Request Throttling Configuration.

        :param policy: The ID of the throttling policy or an instance of
            :class:`~otcextensions.sdk.apig.v2.throttling_policy.
            ThrottlingPolicy`
        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional attributes for the excluded throttling policy
            creation.

        :returns: An instance of ThrottlingExcludedPolicy
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        policy = self._get_resource(_tp.ThrottlingPolicy, policy)
        return self._create(_tx.ThrottlingExcludedPolicy,
                            gateway_id=gateway.id,
                            throttle_id=policy.id,
                            **attrs)

    def update_throttling_excluded_policy(
            self, gateway, policy, exclude, **attrs):
        """Update an Excluded Request Throttling Configuration.

        :param exclude: The ID of the excluded throttling policy or
            an instance of
            :class:`~otcextensions.sdk.apig.v2.throttling_excluded.
            ThrottlingExcludedPolicy`
        :param policy: The ID of the throttling policy or an instance of
            :class:`~otcextensions.sdk.apig.v2.throttling_policy.
            ThrottlingPolicy`
        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional attributes to update the
            excluded throttling policy.

        :returns: Updated instance of ThrottlingExcludedPolicy
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        policy = self._get_resource(_tp.ThrottlingPolicy, policy)
        exclude = self._get_resource(_tx.ThrottlingExcludedPolicy, exclude)
        return self._update(
            _tx.ThrottlingExcludedPolicy,
            exclude,
            gateway_id=gateway.id,
            throttle_id=policy.id,
            **attrs
        )

    def delete_throttling_excluded_policy(
            self, gateway, policy, exclude, ignore_missing=False):
        """Deleting an Excluded Request Throttling Configuration.

        :param exclude: The ID of the excluded throttling policy or
            an instance of
            :class:`~otcextensions.sdk.apig.v2.throttling_excluded.
            ThrottlingExcludedPolicy`
        :param policy: The ID of the throttling policy or an instance of
            :class:`~otcextensions.sdk.apig.v2.throttling_policy.
            ThrottlingPolicy`
        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        policy = self._get_resource(_tp.ThrottlingPolicy, policy)
        exclude = self._get_resource(_tx.ThrottlingExcludedPolicy, exclude)
        return self._delete(
            _tx.ThrottlingExcludedPolicy,
            exclude,
            gateway_id=gateway.id,
            throttle_id=policy.id,
            ignore_missing=ignore_missing
        )

    def throttling_excluded_policies(self, gateway, policy, **query):
        """List all Excluded Request Throttling Configurations.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param policy: The ID of the throttling policy or an instance of
            :class:`~otcextensions.sdk.apig.v2.throttling_policy.
            ThrottlingPolicy`
        :param query: Additional filters for listing excluded throttling
            policies.

        :returns: A list of instances of ThrottlingExcludedPolicy
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        policy = self._get_resource(_tp.ThrottlingPolicy, policy)
        return self._list(
            _tx.ThrottlingExcludedPolicy,
            paginated=False,
            gateway_id=gateway.id,
            throttle_id=policy.id,
            **query
        )

    # ======== Gateway Features Methods ========

    def configure_gateway_feature(self, gateway, **attrs):
        """Configuring a feature for a Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional attributes for the GatewayFeatures.

        :returns: An instance of GatewayFeatures
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._create(
            _gwf.GatewayFeatures,
            gateway_id=gateway.id,
            **attrs)

    def gateway_features(self, gateway, **query):
        """List all Gateway Features.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param query:  Additional filters for listing GatewayFeatures.

        :returns: A list of instances of GatewayFeatures
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _gwf.GatewayFeatures,
            gateway_id=gateway.id,
            **query)

    def supported_gateway_features(self, gateway, **query):
        """List all the supported features of a Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param query:  Additional filters for listing GatewayFeatures.

        :returns: A list of instances of features names
        """

        gateway = self._get_resource(_gateway.Gateway, gateway)
        feat = self._get_resource(_gwf.GatewayFeatures, "")
        return feat._supported_features(self, gateway, **query)

    # ======== Resource Query Methods ========

    def get_api_quantities(self, gateway):
        """Get the number of APIs that have been published in the RELEASE
            environment and the number of APIs that have not been
            published in this environment.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :returns: An instance of ApiQuantities
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._get(
            _rq.ApiQuantities,
            gateway_id=gateway.id,
            requires_id=False,
        )

    def get_api_group_quantities(self, gateway):
        """Get the number of API groups.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :returns: An instance of ApiGroupQuantities
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._get(
            _rq.ApiGroupQuantities,
            gateway_id=gateway.id,
            requires_id=False,
        )

    def get_app_quantities(self, gateway):
        """Get the number of apps that have been authorized to access APIs
            and the number of apps that have not been authorized to access
            any APIs.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :returns: An instance of AppQuantities
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._get(
            _rq.AppQuantities,
            gateway_id=gateway.id,
            requires_id=False,
        )

    # ======== Domain Name Methods ========

    def bind_domain_name(self, gateway, group, **attrs):
        """Bind domain name to group.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param group: The ID of the group or an instance of
            :class:`~otcextensions.sdk.apig.v2.apigroup.ApiGroup`
        :param attrs: Additional attributes for the DomainName.

        :returns: An instance of DomainName
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        group = self._get_resource(_api_group.ApiGroup, group)
        return self._create(
            _domain.DomainName,
            gateway_id=gateway.id,
            group_id=group.id,
            **attrs)

    def unbind_domain_name(
            self, gateway, group, domain, ignore_missing=False):
        """Unbind domain name from group.

        :param domain: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.domain_name.DomainName`
        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param group: The ID of the group or an instance of
            :class:`~otcextensions.sdk.apig.v2.apigroup.ApiGroup`

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        group = self._get_resource(_api_group.ApiGroup, group)
        domain = self._get_resource(_domain.DomainName, domain)
        return self._delete(
            _domain.DomainName,
            domain,
            gateway_id=gateway.id,
            group_id=group.id,
            ignore_missing=ignore_missing
        )

    def update_domain_name_bound(
            self, gateway, group, domain, **attrs):
        """Update a bound of domain name to group.

        :param domain: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.domain_name.DomainName`
        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param group: The ID of the group or an instance of
            :class:`~otcextensions.sdk.apig.v2.apigroup.ApiGroup`
        :param attrs: Additional attributes to update the
            DomainName bind.

        :returns: Updated instance of DomainName
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        group = self._get_resource(_api_group.ApiGroup, group)
        domain = self._get_resource(_domain.DomainName, domain)
        return self._update(
            _domain.DomainName,
            domain,
            gateway_id=gateway.id,
            group_id=group.id,
            **attrs
        )

    def create_certificate_for_domain_name(
            self, gateway, group, domain, **attrs):
        """Add certificate to domain name.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param group: The ID of the group or an instance of
            :class:`~otcextensions.sdk.apig.v2.apigroup.ApiGroup`
        :param domain: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.domain_name.DomainName`
        :param attrs: Additional attributes for the DomainName.

        :returns: An instance of Certificate
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        group = self._get_resource(_api_group.ApiGroup, group)
        domain = self._get_resource(_domain.DomainName, domain)
        return self._create(
            _domain.Certificate,
            gateway_id=gateway.id,
            group_id=group.id,
            domain_id=domain.id,
            **attrs)

    def unbind_certificate_from_domain_name(
            self, gateway, group, domain,
            certificate, ignore_missing=False):
        """Unbind certificate from domain name.

        :param certificate: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.certificate.Certificate`
        :param domain: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.domain_name.DomainName`
        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param group: The ID of the group or an instance of
            :class:`~otcextensions.sdk.apig.v2.apigroup.ApiGroup`
        :param group: The ID of the certificate

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        group = self._get_resource(_api_group.ApiGroup, group)
        domain = self._get_resource(_domain.DomainName, domain)
        certificate = self._get_resource(_c.Certificate, certificate)
        return self._delete(
            _domain.DomainName,
            domain,
            gateway_id=gateway.id,
            group_id=group.id,
            domain_id=domain.id,
            certificate_id=certificate.id,
            ignore_missing=ignore_missing
        )

    def enable_debug_domain_name(self, gateway, group, domain, enable):
        """Disable or Enable the debugging domain name bound to an API group.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param group: The ID of the group or an instance of
            :class:`~otcextensions.sdk.apig.v2.apigroup.ApiGroup`
        :param domain: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.domain_name.DomainName`
        :param enable: Specifies whether the debugging domain name
            is accessible. Options: true and false.

        :returns: An instance of DomainDebug
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        group = self._get_resource(_api_group.ApiGroup, group)
        domain = self._get_resource(_domain.DomainName, domain)
        return self._update(
            _domain.DomainDebug,
            gateway_id=gateway.id,
            group_id=group.id,
            domain_id=domain.id,
            sl_domain_access_enabled=enable)

    def get_bound_certificate(self, gateway, group, domain, certificate):
        """Get the details of the certificate bound to a domain name.

        :param certificate: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.certificate.Certificate`
        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param group: The ID of the group or an instance of
            :class:`~otcextensions.sdk.apig.v2.apigroup.ApiGroup`
        :param domain: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.domain_name.DomainName`

        :returns: An instance of Certificate
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        group = self._get_resource(_api_group.ApiGroup, group)
        domain = self._get_resource(_domain.DomainName, domain)
        certificate = self._get_resource(_c.Certificate, certificate)
        return self._get(
            _domain.Certificate,
            gateway_id=gateway.id,
            group_id=group.id,
            domain_id=domain.id,
            id=certificate.id
        )

    # ======== Certificate Methods ========

    def delete_certificate(self, certificate, ignore_missing=False):
        """Delete an SSL certificate.

        :param certificate: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.certificate.Certificate`

        :returns: None
        """
        certificate = self._get_resource(_c.Certificate, certificate)
        return self._delete(
            _c.Certificate,
            certificate,
            ignore_missing=ignore_missing
        )

    # ======== Credentials Management Methods ========

    def create_app(self, gateway, **attrs):
        """Create a new identity for accessing a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional attributes for the App creation.

        :returns: An instance of App
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._create(_app.App,
                            gateway_id=gateway.id,
                            **attrs)

    def get_app(self, gateway, app):
        """Retrieve details of a specific App.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param app: The ID of the App or an instance of App

        :returns: An instance of App
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        app = self._get_resource(_app.App, app)
        return self._get(
            _app.App,
            app,
            gateway_id=gateway.id
        )

    def update_app(self, gateway, app, **attrs):
        """Update an existing App for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param app: The ID of the App or an instance of App
        :param attrs: Additional attributes to update the App.

        :returns: Updated instance of App
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        app = self._get_resource(_app.App, app)
        return self._update(
            _app.App,
            app,
            gateway_id=gateway.id,
            **attrs
        )

    def delete_app(self, gateway, app, ignore_missing=False):
        """Delete an existing identity from a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param app: The ID of the App or an instance of App

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        app = self._get_resource(_app.App, app)
        return self._delete(
            _app.App,
            app,
            gateway_id=gateway.id,
            ignore_missing=ignore_missing
        )

    def apps(self, gateway, **attrs):
        """Retrieve the list of Apps for a specific API Gateway.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`

        :returns: An instance of App
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _app.App,
            gateway_id=gateway.id,
            **attrs
        )

    def verify_app(self, gateway, app):
        """Verify if the App exists

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param app: The ID of the App or an instance of App

        :returns: An instance of App
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        app = self._get_resource(_app.App, app)
        return app._verify_app(self, gateway)

    def reset_app_secret(self, gateway, app, **attrs):
        """Reset the App secret

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param app: The ID of the App or an instance of App
        :param attrs: Additional attributes to update the App secret.

        :returns: An instance of App
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        app = self._get_resource(_app.App, app)
        return app._reset_secret(self, gateway, **attrs)

    def get_app_code(self, gateway, app, app_code):
        """Retrieve details of a specific application code.

        This method retrieves the details of an application code associated
        with the given API Gateway and application.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param app: The ID of the application or an instance of
            :class:`~otcextensions.sdk.apig.v2.app.App`
        :param app_code: The ID of the application code or an instance of
            :class:`~otcextensions.sdk.apig.v2.app_code.AppCode`

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.app_code.AppCode`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        app = self._get_resource(_app.App, app)
        app_code = self._get_resource(_app_code.AppCode, app_code)
        return self._get(
            _app_code.AppCode,
            app_code,
            gateway_id=gateway.id,
            app_id=app.id,
        )

    def create_app_code(self, gateway, app, **attrs):
        """Create a new application code for a specific application.

        This method creates an application code associated with
        the given API Gateway and application.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param app: The ID of the application or an instance of
            :class:`~otcextensions.sdk.apig.v2.app.App`
        :param attrs: Additional attributes for creating the application code.

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.app_code.AppCode`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        app = self._get_resource(_app.App, app)
        return self._create(
            _app_code.AppCode,
            gateway_id=gateway.id,
            app_id=app.id,
            **attrs
        )

    def generate_app_code(self, gateway, app, **attrs):
        """Generate a new application code for a specific application.

        This method generates a new application code associated with
        the given API Gateway and application.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param app: The ID of the application or an instance of
            :class:`~otcextensions.sdk.apig.v2.app.App`
        :param attrs: Additional attributes for generating the app code.

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.app_code.AppCode`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        app = self._get_resource(_app.App, app)
        app_code = _app_code.AppCode()
        return app_code._generate_app_code(self, gateway, app, **attrs)

    def app_codes(self, gateway, app, **attrs):
        """List all application codes for a specific application.

        This method retrieves a list of application codes associated with
        the given API Gateway and application.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param app: The ID of the application or an instance of
            :class:`~otcextensions.sdk.apig.v2.app.App`
        :param attrs: Additional filters for listing application codes.

        :returns: A list of instances of
            :class:`~otcextensions.sdk.apig.v2.app_code.AppCode`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        app = self._get_resource(_app.App, app)
        return self._list(
            _app_code.AppCode,
            gateway_id=gateway.id,
            app_id=app.id,
            **attrs
        )

    def delete_app_code(self, gateway, app, app_code, ignore_missing=False):
        """Delete a specific application code.

        This method deletes an application code associated with
        the given API Gateway and application.

        :param gateway: The ID of the gateway or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param app: The ID of the application or an instance of
            :class:`~otcextensions.sdk.apig.v2.app.App`
        :param app_code: The ID of the application code or an instance of
            :class:`~otcextensions.sdk.apig.v2.app_code.AppCode`
        :param ignore_missing: When set to True, no exception will be raised
            if the application code does not exist. Default is False.

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        app = self._get_resource(_app.App, app)
        app_code = self._get_resource(_app_code.AppCode, app_code)
        return self._delete(
            _app_code.AppCode,
            app_code,
            gateway_id=gateway.id,
            app_id=app.id,
            ignore_missing=ignore_missing
        )

    def quotas(self, gateway, app, **attrs):
        """Retrieve quotas associated with a credential.

        This method retrieves the quota details associated with
        the given API Gateway instance and application.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param app: The ID of the application or an instance of
            :class:`~otcextensions.sdk.apig.v2.app.App`
        :param attrs: Additional filters for retrieving quota details.

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.quota.Quota`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        app = self._get_resource(_app.App, app)
        return self._get(
            _app.Quota,
            gateway_id=gateway.id,
            app_id=app.id,
            requires_id=False,
            **attrs
        )
    #
    # def access_controls(self, gateway, app, **attrs):
    #     """Retrieve access control details for a specific application.
    #
    #     :param gateway: The ID of the API Gateway instance or an instance of
    #         :class:`~otcextensions.sdk.apig.v2.instance.Instance`
    #     :param app: The ID of the application or an instance of
    #         :class:`~otcextensions.sdk.apig.v2.app.App`
    #     :param attrs: Additional filters for access control details.
    #
    #     :returns: An instance of
    #         :class:`~otcextensions.sdk.apig.v2.access_control.AccessControl`
    #     """
    #     gateway = self._get_resource(_gateway.Gateway, gateway)
    #     app = self._get_resource(_app.App, app)
    #     return self._get(
    #         _ac.AccessControl,
    #         gateway_id = gateway.id,
    #         app_url_id = app.id,
    #         requires_id=False,
    #         **attrs
    #     )
    #
    # def delete_access_control(self, gateway, app, **attrs):
    #     """Delete access control details for a specific application.
    #
    #     :param gateway: The ID of the API Gateway instance or an instance of
    #         :class:`~otcextensions.sdk.apig.v2.instance.Instance`
    #     :param app: The ID of the application or an instance of
    #         :class:`~otcextensions.sdk.apig.v2.app.App`
    #
    #     :returns: None
    #     """
    #     gateway = self._get_resource(_gateway.Gateway, gateway)
    #     app = self._get_resource(_app.App, app)
    #     access_control = _ac.AccessControl()
    #     return access_control._delete(
    #         self,
    #         gateway = gateway,
    #         app = app,
    #         **attrs
    #     )
    #
    # def configure_access_control(self, gateway, app, **attrs):
    #     """Configure access control details for a specific application.
    #
    #     :param gateway: The ID of the API Gateway instance or an instance of
    #         :class:`~otcextensions.sdk.apig.v2.instance.Instance`
    #     :param app: The ID of the application or an instance of
    #         :class:`~otcextensions.sdk.apig.v2.app.App`
    #     :param attrs: Additional attributes for configuring access control.
    #
    #     :returns: An instance of
    #         :class:`~otcextensions.sdk.apig.v2.access_control.AccessControl`
    #     """
    #     gateway = self._get_resource(_gateway.Gateway, gateway)
    #     app = self._get_resource(_app.App, app)
    #     access_control = _ac.AccessControl()
    #     return access_control._configure(self, gateway, app, **attrs)

    # ======== App Authorization Methods ========

    def list_api_bound_to_app(self, gateway, **attrs):
        """List all APIs authorized (bound) to a specific application.

        This method retrieves a list of APIs that are bound (authorized)
        to the given application within the specified API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional filters, such as app_id, env_id, or API name.

        :returns: A list of instances of
            :class:`~otcextensions.sdk.apig.v2.auth.ApiAuthInfo`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _auth.ApiAuthInfo,
            paginated=False,
            base_path=f'{_auth.ApiAuthInfo.base_path}/binded-apis',
            gateway_id=gateway.id,
            **attrs
        )

    def list_apps_bound_to_api(self, gateway, **attrs):
        """List all applications authorized (bound) to a specific API.

        This method gets a list of applications that are bound (authorized)
        to the given API within the specified API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional filters, such as api_id or environment ID.

        :returns: A list of instances of
            :class:`~otcextensions.sdk.apig.v2.auth.ApiAuthInfo`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _auth.ApiAuthInfo,
            paginated=False,
            base_path=f'{_auth.ApiAuthInfo.base_path}/binded-apps',
            gateway_id=gateway.id,
            **attrs
        )

    def list_api_not_bound_to_app(self, gateway, **attrs):
        """List all APIs not authorized (not bound) to a specific application.

        This method retrieves a list of APIs that are not bound (unauthorized)
        to the given application within the specified API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional filters, such as app_id or environment ID.

        :returns: A list of instances of
            :class:`~otcextensions.sdk.apig.v2.auth.ApiAuth`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _auth.ApiAuth,
            paginated=False,
            base_path=f'{_auth.ApiAuthInfo.base_path}/unbinded-apis',
            gateway_id=gateway.id,
            **attrs
        )

    def create_auth_in_api(self, gateway, **attrs):
        """Authorize one or more applications to access a specific API.

        This method binds applications to the specified API within the given
        API Gateway instance, effectively authorizing them to access the API.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Attributes required for authorization, including:
            - api_id: ID of the API to authorize.
            - app_ids: List of application IDs to bind to the API.
            - env_id: ID of the environment in which the API is published.

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.auth.ApiAuthInfo`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        auth = _auth.ApiAuthInfo()
        return auth._authorize_apps(
            self,
            gateway_id=gateway.id,
            **attrs
        )

    def delete_auth_from_api(self, gateway, auth_id):
        """Delete an API authorization from an application.

        This method removes the authorization binding between a specific
        API and an application within the given API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param auth_id: The ID of the authorization binding to be deleted.

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        auth = _auth.ApiAuthInfo()
        return auth._cancel_auth(
            self,
            gateway_id=gateway.id,
            app_auth_id=auth_id
        )

    # ======== Access Control Policy Methods ========

    def create_acl_policy(self, gateway, **attrs):
        """Create an access control policy.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Attributes required to create the ACL policy

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.acl_policy.AclPolicy`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._create(
            _acl.AclPolicy,
            gateway_id=gateway.id,
            **attrs
        )

    def update_acl_policy(self, gateway, acl_policy, **attrs):
        """Update an existing access control policy.

        This method updates an existing ACL (access control list) policy
        in the specified API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param acl_policy: The ID of the ACL policy or an instance of
            :class:`~otcextensions.sdk.apig.v2.acl_policy.AclPolicy`
        :param attrs: Attributes to update

        :returns: The updated instance of
            :class:`~otcextensions.sdk.apig.v2.acl_policy.AclPolicy`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        acl_policy = self._get_resource(_acl.AclPolicy, acl_policy)
        return self._update(
            _acl.AclPolicy,
            acl_policy,
            gateway_id=gateway.id,
            **attrs
        )

    def delete_acl_policy(self, gateway, acl_policy, ignore_missing=True,
                          **attrs):
        """Delete an access control policy.

        This method deletes an existing access control (ACL) policy from
        the specified API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param acl_policy: The ID of the ACL policy or an instance of
            :class:`~otcextensions.sdk.apig.v2.acl_policy.AclPolicy`
        :param ignore_missing: If True, no exception is raised when the ACL
            policy does not exist
        :param attrs: Additional attributes for the delete operation

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        acl_policy = self._get_resource(_acl.AclPolicy, acl_policy)
        return self._delete(
            _acl.AclPolicy,
            acl_policy,
            gateway_id=gateway.id,
            ignore_missing=ignore_missing,
            **attrs
        )

    def delete_acl_policies(self, gateway, **attrs):
        """Delete multiple access control policies in batch.

        This method deletes multiple ACL (access control list) policies at once
        within the specified API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Attributes for batch deletion

        :returns: A response indicating the result of the batch deletion.
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        acl = _acl.AclPolicy()
        return acl._delete_multiple_acls(
            self,
            gateway_id=gateway.id,
            **attrs
        )

    def acl_policies(self, gateway, **attrs):
        """List all access control policies in an API Gateway instance.

        This method retrieves a list of all ACL (access control list) policies
        defined in the specified API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional filters for listing ACL policies

        :returns: A list of instances of
            :class:`~otcextensions.sdk.apig.v2.acl_policy.AclPolicy`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _acl.AclPolicy,
            paginated=False,
            gateway_id=gateway.id,
            **attrs
        )

    def get_acl_policy(self, gateway, acl_policy, **attrs):
        """Retrieve details of a specific access control policy.

        This method retrieves detailed information about an existing ACL
        policy within the specified API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param acl_policy: The ID of the ACL policy or an instance of
            :class:`~otcextensions.sdk.apig.v2.acl_policy.AclPolicy`
        :param attrs: Additional parameters for retrieving the ACL policy.

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.acl_policy.AclPolicy`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        acl_policy = self._get_resource(_acl.AclPolicy, acl_policy)
        return self._get(
            _acl.AclPolicy,
            acl_policy,
            gateway_id=gateway.id,
            **attrs
        )

    # ======== Binding/Unbinding Access Control Policies Methods ========

    def list_apis_for_acl(self, gateway, **attrs):
        """List all APIs bound to a specific access control policy.

        This method retrieves a list of APIs that are associated with
        the specified ACL policy in the given API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional filters

        :returns: A list of instances of
            :class:`~otcextensions.sdk.apig.v2.acl_api_binding.ApiForAcl`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _acl_api_binding.ApiForAcl,
            paginated=False,
            gateway_id=gateway.id,
            **attrs
        )

    def list_api_not_bound_to_acl(self, gateway, **attrs):
        """List all APIs not bound to a specific access control policy.

        This method retrieves a list of APIs that are not associated with
        the specified ACL policy in the given API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional filters

        :returns: A list of instances of
            :class:`~otcextensions.sdk.apig.v2.acl_api_binding.UnbindApiForAcl`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _acl_api_binding.UnbindApiForAcl,
            paginated=False,
            gateway_id=gateway.id,
            **attrs
        )

    def list_acl_for_api(self, gateway, **attrs):
        """List all access control policies bound to a specific API.

        This method retrieves a list of ACL policies that are associated with
        the specified API in the given API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional filters

        :returns: A list of instances of
            :class:`~otcextensions.sdk.apig.v2.acl_api_binding.AclForApi`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _acl_api_binding.AclForApi,
            paginated=False,
            gateway_id=gateway.id,
            **attrs
        )

    def bind_acl_to_api(self, gateway, **attrs):
        """Bind an access control policy to one or more APIs.

        This method binds an existing ACL policy to one
        or more APIs within the specified API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Attributes for the binding

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.acl_api_binding.AclApiBinding`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        acl = _acl_api_binding.AclApiBinding()
        return acl._bind_to_api(
            self,
            gateway_id=gateway.id,
            **attrs
        )

    def unbind_acl(self, gateway, acl, ignore_missing=True):
        """Unbind an access control policy from an API.

        This method removes the binding between an ACL (access control list)
        policy and an API within the specified API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param acl: The ID of the ACL binding or an instance of
            :class:`~otcextensions.sdk.apig.v2.acl_api_binding.AclApiBinding`
        :param ignore_missing: If True, no exception is raised if the binding
            does not exist

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        acl = self._get_resource(_acl_api_binding.AclApiBinding, acl)
        return self._delete(
            _acl_api_binding.AclApiBinding,
            acl,
            gateway_id=gateway.id,
            ignore_missing=ignore_missing
        )

    def unbind_acls(self, gateway, **attrs):
        """Unbind multiple access control policies from APIs in batch.

        This method removes bindings between one or more ACL policies
        and APIs within the specified API Gateway instance

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Attributes for the unbinding operation

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.acl_api_binding.
            AclBindingFailure`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        acl = _acl_api_binding.AclBindingFailure()
        return acl._unbind_multiple_acls(
            self,
            gateway_id=gateway.id,
            **attrs
        )

    # ======== Custom Authorizer Methods ========

    def custom_authorizers(self, gateway, **attrs):
        """List all custom authorizers in an API Gateway instance

        This method retrieves a list of custom authorizers configured within
        the specified API Gateway instance. Custom authorizers allow you to
        define your own logic for validating access tokens and identities

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional filters for listing custom authorizers,
            such as name or type

        :returns: A list of instances of
            :class:`~otcextensions.sdk.apig.v2.custom_authorizer.
            CustomAuthorizer`
            """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _custom_auth.CustomAuthorizer,
            paginated=False,
            gateway_id=gateway.id,
            **attrs
        )

    def get_custom_authorizer(self, gateway, custom_authorizer, **attrs):
        """Retrieve details of a specific custom authorizer

        This method retrieves detailed information about a custom authorizer
        within the specified API Gateway instance

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param custom_authorizer: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.custom_authorizer
            CustomAuthorizer`
        :param attrs: Additional parameters for retrieving
            the custom authorizer

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.custom_authorizer.
            CustomAuthorizer`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        custom_authorizer = self._get_resource(_custom_auth.CustomAuthorizer,
                                               custom_authorizer)
        return self._get(
            _custom_auth.CustomAuthorizer,
            custom_authorizer,
            gateway_id=gateway.id,
            **attrs
        )

    def create_custom_authorizer(self, gateway, **attrs):
        """Create a custom authorizer in an API Gateway instance

            This method creates a new custom authorizer within the specified
            API Gateway instance. Custom authorizers enable custom
            authentication and authorization logic for APIs

            :param gateway: The ID of the API Gateway or an instance of
                :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
            :param attrs: Attributes required to create the custom authorizer

            :returns: An instance of
                :class:`~otcextensions.sdk.apig.v2.custom_authorizer.
                CustomAuthorizer`
            """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._create(
            _custom_auth.CustomAuthorizer,
            gateway_id=gateway.id,
            **attrs
        )

    def update_custom_authorizer(self, gateway, custom_authorizer, **attrs):
        """Update a custom authorizer

        This method updates an existing custom authorizer within the specified
        API Gateway instance

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param custom_authorizer: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.custom_authorizer.
            CustomAuthorizer`
        :param attrs: Attributes to update

        :returns: The updated instance of
            :class:`~otcextensions.sdk.apig.v2.custom_authorizer.
            CustomAuthorizer`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        custom_authorizer = self._get_resource(_custom_auth.CustomAuthorizer,
                                               custom_authorizer)
        return self._update(
            _custom_auth.CustomAuthorizer,
            custom_authorizer,
            gateway_id=gateway.id,
            **attrs
        )

    def delete_custom_authorizer(self, gateway, custom_authorizer,
                                 ignore_missing=False,
                                 **attrs):
        """Delete a custom authorizer

        This method deletes a custom authorizer from the specified
        API Gateway instance

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param custom_authorizer: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.custom_authorizer.
            CustomAuthorizer`
        :param ignore_missing: If True, no exception is raised if
            the authorizer does not exist
        :param attrs: Additional parameters for the delete operation

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        custom_authorizer = self._get_resource(_custom_auth.CustomAuthorizer,
                                               custom_authorizer)
        return self._delete(
            _custom_auth.CustomAuthorizer,
            custom_authorizer,
            gateway_id=gateway.id,
            ignore_missing=ignore_missing,
            **attrs
        )

    def import_api(self, gateway, **attrs):
        """Import an API definition into API Gateway

        This method imports an API from an OpenAPI/Swagger definition or
        other supported format into the specified API Gateway instance.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Dictionary of parameters required for import.

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.export_import.ImportApi`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        import_api = _export_api.ImportApi()
        return import_api._import_api(
            self,
            gateway_id=gateway.id,
            **attrs
        )

    def export_api(self, gateway, full_path, **attrs):
        """Export an API definition from API Gateway

        This method exports an API definition (e.g. in Swagger/OpenAPI format)
        from the specified API Gateway instance to a file.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param full_path: Path where the exported file will be stored
        :param attrs: Optional parameters for export.

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.export_import.ExportApi`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        export_api = _export_api.ExportApi()
        return export_api._export_api(
            self,
            gateway_id=gateway.id,
            full_path=full_path,
            **attrs
        )
    # ======== VPC Channel Methods ========

    def create_vpc_channel(self, gateway, **attrs):
        """Create a VPC channel under the specified API Gateway instance

        This method creates a new VPC channel, which defines how the
        gateway communicates with backend services.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Dictionary of parameters for VPC channel creation.

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._create(
            _vpc_channel.VpcChannel,
            gateway_id=gateway.id,
            **attrs
        )

    def update_vpc_channel(self, gateway, vpc_channel, **attrs):
        """Update an existing VPC channel

        This method updates the properties of a VPC channel under the
        specified API Gateway instance.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        :param attrs: Dictionary of attributes to update on the VPC channel

        :returns: An updated instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        vpc_channel = self._get_resource(_vpc_channel.VpcChannel, vpc_channel)
        return self._update(
            _vpc_channel.VpcChannel,
            vpc_channel,
            gateway_id=gateway.id,
            **attrs
        )

    def delete_vpc_channel(self, gateway, vpc_channel, ignore_missing=False):
        """Delete a VPC channel

        This method deletes the specified VPC channel from the given
        API Gateway instance.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        :param ignore_missing: When set to True, no exception will be raised
            if the application code does not exist. Default is False.

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        vpc_channel = self._get_resource(_vpc_channel.VpcChannel, vpc_channel)
        return self._delete(
            _vpc_channel.VpcChannel,
            vpc_channel,
            gateway_id=gateway.id,
            ignore_missing=ignore_missing
        )

    def get_vpc_channel(self, gateway, vpc_channel):
        """Retrieve details of a VPC channel

        This method fetches the details of the specified VPC channel
        under the given API Gateway instance.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._get(_vpc_channel.VpcChannel,
                         vpc_channel,
                         gateway_id=gateway.id)

    def vpc_channels(self, gateway, **attrs):
        """List all VPC channels under a gateway

        This method retrieves all VPC channels configured for the specified
        API Gateway instance.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Optional query parameters for filtering the list

        :returns: A generator of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        return self._list(
            _vpc_channel.VpcChannel,
            gateway_id=gateway.id,
            **attrs
        )

    def modify_vpc_channel_healthcheck(self, gateway, vpc_channel, **attrs):
        """Update the health check configuration of a VPC channel

        This method modifies the health check settings (e.g. path, interval,
        timeout, thresholds) for the specified VPC channel.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        :param attrs: Dictionary of health check configuration parameters.

        :returns: An updated instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        vpc_channel = self._get_resource(_vpc_channel.VpcChannel, vpc_channel)
        return vpc_channel.modify_healthcheck(
            self,
            gateway_id=gateway.id,
            vpc_channel_id=vpc_channel.id,
            **attrs
        )

    def add_or_update_backend_server_group(self, gateway, vpc_channel,
                                           **attrs):
        """Add or update a backend server group for a VPC channel

        This method creates or updates a backend server group within the
        specified VPC channel of an API Gateway instance.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        :param attrs: Dictionary of parameters for creation or update.

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.backend_server_group.
            BackendServerGroup`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        vpc_channel = self._get_resource(_vpc_channel.VpcChannel, vpc_channel)
        member_group = _backend_group.BackendServerGroup()
        return member_group.create_group(
            self,
            gateway_id=gateway.id,
            vpc_channel_id=vpc_channel.id,
            **attrs
        )

    def backend_server_groups(self, gateway, vpc_channel, **attrs):
        """List backend server groups under a VPC channel

        This method retrieves all backend server groups configured within the
        specified VPC channel of an API Gateway instance.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        :param attrs: Optional query parameters for filtering the list

        :returns: A generator of
            :class:`~otcextensions.sdk.apig.v2.backend_server_group.
            BackendServerGroup` instances
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        vpc_channel = self._get_resource(_vpc_channel.VpcChannel, vpc_channel)
        return self._list(
            _backend_group.BackendServerGroup,
            gateway_id=gateway.id,
            vpc_channel_id=vpc_channel.id,
            **attrs
        )

    def get_backend_server_group(self, gateway, vpc_channel, backend_group):
        """Retrieve details of a backend server group

        This method fetches details of the specified backend server group
        under a given VPC channel of an API Gateway instance.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        :param backend_group: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.backend_server_group.
            BackendServerGroup`

        :returns: An instance of
            :class:`~otcextensions.sdk.apig.v2.backend_server_group.
            BackendServerGroup`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        vpc_channel = self._get_resource(_vpc_channel.VpcChannel, vpc_channel)
        backend_group = self._get_resource(_backend_group.BackendServerGroup,
                                           backend_group)
        return self._get(
            _backend_group.BackendServerGroup,
            backend_group.member_group_id,
            gateway_id=gateway.id,
            vpc_channel_id=vpc_channel.id
        )

    def update_backend_server_group(self, gateway, vpc_channel, backend_group,
                                    **attrs):
        """Update a backend server group

        This method updates the specified backend server group under a given
        VPC channel of an API Gateway instance.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        :param backend_group: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.backend_server_group.
            BackendServerGroup`
        :param attrs: Dictionary of attributes to update on the backend group.

        :returns: An updated instance of
            :class:`~otcextensions.sdk.apig.v2.backend_server_group.
            BackendServerGroup`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        vpc_channel = self._get_resource(_vpc_channel.VpcChannel, vpc_channel)
        backend_group = self._get_resource(_backend_group.BackendServerGroup,
                                           backend_group)
        return self._update(
            _backend_group.BackendServerGroup,
            backend_group.member_group_id,
            gateway_id=gateway.id,
            vpc_channel_id=vpc_channel.id,
            **attrs
        )

    def delete_backend_server_group(self, gateway, vpc_channel, backend_group,
                                    ignore_missing=False):
        """Delete a backend server group

        This method deletes the specified backend server group from a VPC
        channel under an API Gateway instance.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        :param backend_group: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.backend_server_group.
            BackendServerGroup`
        :param ignore_missing: When set to True, no exception will be raised
            if the application code does not exist. Default is False.

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        vpc_channel = self._get_resource(_vpc_channel.VpcChannel, vpc_channel)
        backend_group = self._get_resource(_backend_group.BackendServerGroup,
                                           backend_group)
        self._delete(
            _backend_group.BackendServerGroup,
            backend_group.member_group_id,
            gateway_id=gateway.id,
            vpc_channel_id=vpc_channel.id,
            ignore_missing=ignore_missing
        )

    def add_or_update_backend_servers(self, gateway, vpc_channel, **attrs):
        """Add or update backend server instances

        This method adds or updates backend server instances directly
        under a VPC channel of the specified API Gateway instance.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        :param attrs: Dictionary containing list of members to add or update.

        :returns: A list of
            :class:`~otcextensions.sdk.apig.v2.backend_server.BackendServer`
            instances
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        vpc_channel = self._get_resource(_vpc_channel.VpcChannel, vpc_channel)
        member = _backend_server.BackendServer()
        return member.create_members(
            self,
            gateway_id=gateway.id,
            vpc_channel_id=vpc_channel.id,
            **attrs
        )

    def list_backend_servers(self, gateway, vpc_channel, **attrs):
        """List backend server instances

        This method retrieves all backend server instances configured
        under a specified VPC channel in an API Gateway instance.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        :param attrs: Optional query parameters for filtering

        :returns: A generator of
            :class:`~otcextensions.sdk.apig.v2.backend_server.BackendServer`
            instances
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        vpc_channel = self._get_resource(_vpc_channel.VpcChannel, vpc_channel)
        return self._list(
            _backend_server.BackendServer,
            gateway_id=gateway.id,
            vpc_chan_id=vpc_channel.id,
            **attrs
        )

    def update_backend_server(self, gateway, vpc_channel, **attrs):
        """Update backend server instance(s)

        This method updates one or more backend server instances
        under a specified VPC channel.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        :param attrs: Dictionary with the list of members to update.

        :returns: A list of
            :class:`~otcextensions.sdk.apig.v2.backend_server.BackendServer`
            instances
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        vpc_channel = self._get_resource(_vpc_channel.VpcChannel, vpc_channel)
        member = _backend_server.BackendServer()
        return member.update_members(
            self,
            gateway_id=gateway.id,
            vpc_channel_id=vpc_channel.id,
            **attrs
        )

    def remove_backend_server(self, gateway, vpc_channel, backend_server,
                              ignore_missing=False):
        """Delete a backend server instance

        This method deletes a specific backend server from a VPC channel
        under the specified API Gateway instance.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        :param backend_server: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.backend_server.BackendServer`
        :param ignore_missing: When set to True, no exception will be raised
            if the application code does not exist. Default is False.

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        vpc_channel = self._get_resource(_vpc_channel.VpcChannel, vpc_channel)
        backend_server = self._get_resource(_backend_server.BackendServer,
                                            backend_server)
        self._delete(
            _backend_server.BackendServer,
            backend_server,
            gateway_id=gateway.id,
            vpc_chan_id=vpc_channel.id,
            ignore_missing=ignore_missing
        )

    def enable_backend_server(self, gateway, vpc_channel, backend_server,
                              **attrs):
        """Enable a backend server instance

        This method enables a specific backend server instance by sending
        a batch-enable request with the server's ID.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        :param backend_server: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.backend_server.BackendServer`
        :param attrs: Optional parameters for the enable operation.

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        vpc_channel = self._get_resource(_vpc_channel.VpcChannel, vpc_channel)
        backend_server = self._get_resource(_backend_server.BackendServer,
                                            backend_server)
        backend_server.enable_server(
            self,
            gateway_id=gateway.id,
            vpc_chan_id=vpc_channel.id,
            **attrs
        )

    def disable_backend_server(self, gateway, vpc_channel, backend_server,
                               **attrs):
        """Disable a backend server instance

        This method disables a specific backend server instance by sending
        a batch-disable request with the server's ID.

        :param gateway: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param vpc_channel: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.vpc_channel.VpcChannel`
        :param backend_server: The ID or an instance of
            :class:`~otcextensions.sdk.apig.v2.backend_server.BackendServer`
        :param attrs: Optional parameters for the disable operation.

        :returns: None
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        vpc_channel = self._get_resource(_vpc_channel.VpcChannel, vpc_channel)
        backend_server = self._get_resource(_backend_server.BackendServer,
                                            backend_server)
        backend_server.disable_server(
            self,
            gateway_id=gateway.id,
            vpc_chan_id=vpc_channel.id,
            **attrs
        )

    # ======== Monitoring Information Query Methods ========

    def list_api_calls_for_period(self, gateway, **attrs):
        """List API call statistics for a specified period

        This method retrieves API call statistics for the specified
        API Gateway instance over a defined time period.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional parameters for filtering the results,
            such as start and end time, API ID, etc.

        :returns: A list of instances of
            :class:`~otcextensions.sdk.apig.v2.api_call.ApiCallResult`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        api_call = _api_call.ApiCallResult()
        return api_call.get_api_calls_for_period(
            self,
            gateway_id=gateway.id,
            **attrs
        )

    def list_api_calls_for_group(self, gateway, **attrs):
        """List API call statistics for a specific group

        This method retrieves API call statistics for a specific group
        within the specified API Gateway instance.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional parameters for filtering the results,
            such as group ID, start and end time, etc.

        :returns: A list of instances of
            :class:`~otcextensions.sdk.apig.v2.api_call.ApiCallResult`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        api_call = _api_call.ApiCallResult()
        return api_call.get_api_calls_for_group(
            self,
            gateway_id=gateway.id,
            **attrs
        )

    def list_metric_data(self, gateway, **attrs):
        """List metric data for the API Gateway instance

        This method retrieves metric data for the specified API Gateway
        instance, allowing you to monitor performance and usage statistics.

        :param gateway: The ID of the API Gateway instance or an instance of
            :class:`~otcextensions.sdk.apig.v2.gateway.Gateway`
        :param attrs: Additional parameters for filtering the results,
            such as metric name, time range, etc.

        :returns: A list of instances of
            :class:`~otcextensions.sdk.apig.v2.metric_data.MetricData`
        """
        gateway = self._get_resource(_gateway.Gateway, gateway)
        metric_data = _metric_data.MetricData()
        return metric_data.get_metric_data(
            self,
            gateway_id=gateway.id,
            **attrs
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import urllib

from openstack import resource
from openstack import exceptions


class Export(resource.Resource):
    base_path = '/fgs/functions/%(function_urn)s/export'

    # Capabilities
    allow_fetch = True

    _query_mapping = resource.QueryParameters(
        'config', 'code', 'type'
    )

    # Properties
    function_urn = resource.URI('function_urn', type=str)

    # Attributes
    file_name = resource.Body('file_name', type=str)
    code_url = resource.Body('code_url', type=str)

    def _export(self, session, function, **query):
        """Export a function
        """
        query_params = urllib.parse.urlencode(query)
        urn = function.func_urn.rpartition(":")[0]
        url = self.base_path % {'function_urn': urn}
        url += '?' + query_params
        response = session.get(url)
        exceptions.raise_from_response(response)
        self._translate_response(response)
        return self



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack import resource


class Server(resource.Resource):
    resource_key = ''
    resources_key = 'servers'
    base_path = '/dedicated-hosts/%(dedicated_host_id)s/servers'

    # capabilities
    allow_list = True

    #: Properties
    addresses = resource.Body('addresses', type=dict)
    created_at = resource.Body('created')
    dedicated_host_id = resource.URI('dedicated_host_id')
    flavor = resource.Body('flavor')
    metadata = resource.Body('metadata')
    status = resource.Body('status')
    tenant_id = resource.Body('tenant_id')
    updated_at = resource.Body('updated')
    user_id = resource.Body('user_id')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import uuid

import fixtures

from openstackclient.tests.functional import base


CREATE_COMMAND = 'obs container create %(name)s -f json'

DELETE_COMMAND = 'obs container delete %(name)s'


class ObsContainerTests(base.TestCase):
    """Functional tests for vbs. """

    NAME = uuid.uuid4().hex
    OTHER_NAME = uuid.uuid4().hex

    def tearDown(self):
        try:
            self.openstack(
                DELETE_COMMAND % {'name': self.NAME}
            )
        finally:
            super(ObsContainerTests, self).tearDown()

    def setUp(self):
        super(ObsContainerTests, self).setUp()
        json_output = json.loads(self.openstack(
            CREATE_COMMAND % {'name': self.NAME}
        ))
        self.container_id = json_output["id"]
        self.assertOutput(self.NAME, json_output['name'])

        ver_fixture = fixtures.EnvironmentVariable(
            'OS_OBS_API_VERSION', '1'
        )
        self.useFixture(ver_fixture)

    def test_container_list(self):
        json_output = json.loads(self.openstack(
            'obs container list -f json '
        ))
        self.assertIn(
            self.NAME,
            [cont['name'] for cont in json_output]
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
'''LoadBalancer Health Monitor v1 action implementations'''
import logging

from osc_lib import utils
from osc_lib.command import command

from otcextensions.common import sdk_utils
from otcextensions.i18n import _

LOG = logging.getLogger(__name__)


_formatters = {
    'pool_ids': sdk_utils.ListOfIdsColumnBR,
}


def _get_columns(item):
    column_map = {
        'is_admin_state_up': 'admin_state_up',
    }
    return sdk_utils.get_osc_show_columns_for_sdk_resource(item, column_map)


HTTP_METHODS = ['GET', 'HEAD', 'POST', 'PUT', 'DELETE',
                'TRACE', 'OPTIONS', 'CONNECT', 'PATCH']
TYPE_VALUES = ['HTTP', 'HTTPS', 'PING', 'TCP', 'TLS-HELLO', 'UDP_CONNECT']


class ListHealthMonitor(command.Lister):
    _description = _('List health monitors')
    column_headers = ('id', 'name', 'project_id', 'type', 'admin_state_up')
    columns = ('id', 'name', 'project_id', 'type', 'is_admin_state_up')

    def get_parser(self, prog_name):
        parser = super(ListHealthMonitor, self).get_parser(prog_name)

        parser.add_argument(
            '--type',
            metavar='{' + ','.join(TYPE_VALUES) + '}',
            type=lambda s: s.upper(),
            choices=TYPE_VALUES,
            help=_('Health monitor type to use as a filter\n'
                   'one of [`HTTP`, `HTTPS`, `PING`, `TCP`, `TLS-HELLO`, '
                   '`UDP_CONNECT`]')
        )

        return parser

    def take_action(self, parsed_args):

        args = {}
        if parsed_args.type:
            args['type'] = parsed_args.type

        client = self.app.client_manager.network

        data = client.health_monitors(**args)

        return (
            self.column_headers,
            (utils.get_item_properties(
                s, self.columns, formatters=_formatters
            ) for s in data))


class ShowHealthMonitor(command.ShowOne):
    _description = _('Show the details of a single health monitor')

    def get_parser(self, prog_name):
        parser = super(ShowHealthMonitor, self).get_parser(prog_name)

        parser.add_argument(
            'health_monitor',
            metavar='<health_monitor>',
            help=_('Name or UUID of the health monitor.')
        )

        return parser

    def take_action(self, parsed_args):

        args = {}

        client = self.app.client_manager.network

        obj = client.find_health_monitor(
            name_or_id=parsed_args.health_monitor,
            ignore_missing=False,
            **args)

        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters=_formatters)

        return (display_columns, data)


class CreateHealthMonitor(command.ShowOne):
    _description = _('Create a health monitor')

    def get_parser(self, prog_name):
        parser = super(CreateHealthMonitor, self).get_parser(prog_name)

        parser.add_argument(
            'pool',
            metavar='<pool>',
            help=_('Set the pool for the health monitor (name or ID).')
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Set the health monitor name.')
        )
        parser.add_argument(
            '--delay',
            metavar='<delay>',
            type=int,
            required=True,
            help=_('Set the time in seconds, between sending probes '
                   'to members.')
        )
        parser.add_argument(
            '--expected_codes',
            metavar='<codes>',
            help=_('The list of HTTP status codes expected in response '
                   'from the member to declare it healthy. Specify one '
                   'of the following values:\n'
                   '* A single value, such as 200\n'
                   '* A list, such as 200, 202\n'
                   '* A range, such as 200-204\n'
                   'The default is 200.')
        )
        parser.add_argument(
            '--http_method',
            metavar='{' + ','.join(HTTP_METHODS) + '}',
            choices=HTTP_METHODS,
            type=lambda s: s.upper(),  # case insensitive
            help=_('The HTTP method that the health monitor uses for requests.'
                   'One of [`GET`, `HEAD`, `POST`, `PUT`, `DELETE`,'
                   '`TRACE`, `OPTIONS`, `CONNECT`, `PATCH`]\n'
                   'Default is `GET`')
        )
        parser.add_argument(
            '--timeout',
            metavar='<timeout>',
            type=int,
            help=_('The maximum time, in seconds, that a monitor waits to '
                   'connect before it times out. This value must be less '
                   'than the delay value.')
        )
        parser.add_argument(
            '--max_retries',
            metavar='<max_retries>',
            type=int,
            required=True,
            choices=range(1, 10),
            help=_('The number of successful checks before changing the '
                   'operating status of the member to `ONLINE`. '
                   'A valid value is from 1 to 10.')
        )
        parser.add_argument(
            '--url_path',
            metavar='<url_path>',
            help=_('The HTTP URL path of the request sent by the monitor '
                   'to test the health of a backend member. Must be a string '
                   'that begins with a forward slash (/).'
                   'The default URL path is /.')
        )
        parser.add_argument(
            '--type',
            metavar='{' + ','.join(TYPE_VALUES) + '}',
            choices=TYPE_VALUES,
            type=lambda s: s.upper(),  # case insensitive
            required=True,
            help=_('The type of health monitor.\n'
                   'one of [`HTTP`, `HTTPS`, `PING`, `TCP`, `TLS-HELLO`, '
                   '`UDP_CONNECT`]')
        )
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable',
            action='store_true',
            default=True,
            help=_('Enable health monitor (default).')
        )
        admin_group.add_argument(
            '--disable',
            action='store_true',
            default=None,
            help=_('Disable health monitor.')
        )

        return parser

    def take_action(self, parsed_args):

        args = {}
        if parsed_args.delay:
            args['delay'] = parsed_args.delay
        if parsed_args.expected_codes:
            args['expected_codes'] = parsed_args.expected_codes
        if parsed_args.max_retries:
            args['max_retries'] = parsed_args.max_retries
        if parsed_args.timeout:
            args['timeout'] = parsed_args.timeout
        if parsed_args.url_path:
            args['url_path'] = parsed_args.url_path
        if parsed_args.http_method:
            args['http_method'] = parsed_args.http_method
        if parsed_args.type:
            args['type'] = parsed_args.type
        if parsed_args.name:
            args['name'] = parsed_args.name
        if parsed_args.disable:
            args['is_admin_state_up'] = False

        client = self.app.client_manager.network

        pool = client.find_pool(name_or_id=parsed_args.pool,
                                ignore_missing=False)

        args['pool_id'] = pool.id

        obj = client.create_health_monitor(
            **args)

        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters=_formatters)

        return (display_columns, data)


class SetHealthMonitor(command.ShowOne):
    _description = _('Update a health monitor')

    def get_parser(self, prog_name):
        parser = super(SetHealthMonitor, self).get_parser(prog_name)

        parser.add_argument(
            'health_monitor',
            metavar='<health_monitor>',
            help=_('Health monitor to update (name or ID).')
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Human-readable name of the resource.')
        )
        parser.add_argument(
            '--delay',
            metavar='<delay>',
            type=int,
            help=_('The time, in seconds, between sending probes to members.')
        )
        parser.add_argument(
            '--expected_codes',
            metavar='<expected_codes>',
            help=_('The list of HTTP status codes expected in response '
                   'from the member to declare it healthy. Specify one '
                   'of the following values:\n'
                   '* A single value, such as 200\n'
                   '* A list, such as 200, 202\n'
                   '* A range, such as 200-204\n'
                   'The default is 200.')
        )
        parser.add_argument(
            '--http_method',
            metavar='{' + ','.join(HTTP_METHODS) + '}',
            choices=HTTP_METHODS,
            type=lambda s: s.upper(),  # case insensitive
            help=_('The HTTP method that the health monitor uses for requests.'
                   'One of [`GET`, `HEAD`, `POST`, `PUT`, `DELETE`,'
                   '`TRACE`, `OPTIONS`, `CONNECT`, `PATCH`]\n'
                   'Default is `GET`')
        )
        parser.add_argument(
            '--timeout',
            metavar='<timeout>',
            type=int,
            help=_('The maximum time, in seconds, that a monitor waits to '
                   'connect before it times out. This value must be less '
                   'than the delay value.')
        )
        parser.add_argument(
            '--max_retries',
            metavar='<max_retries>',
            type=int,
            required=True,
            choices=range(1, 10),
            help=_('The number of successful checks before changing the '
                   'operating status of the member to `ONLINE`. '
                   'A valid value is from 1 to 10.')
        )
        parser.add_argument(
            '--url_path',
            metavar='<url_path>',
            help=_('The HTTP URL path of the request sent by the monitor '
                   'to test the health of a backend member. Must be a string '
                   'that begins with a forward slash (/).'
                   'The default URL path is /.')
        )
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable',
            action='store_true',
            default=True,
            help=_('Enable health monitor (default).')
        )
        admin_group.add_argument(
            '--disable',
            action='store_true',
            default=None,
            help=_('Disable health monitor.')
        )

        return parser

    def take_action(self, parsed_args):

        args = {}
        if parsed_args.delay:
            args['delay'] = parsed_args.delay
        if parsed_args.expected_codes:
            args['expected_codes'] = parsed_args.expected_codes
        if parsed_args.max_retries:
            args['max_retries'] = parsed_args.max_retries
        if parsed_args.timeout:
            args['timeout'] = parsed_args.timeout
        if parsed_args.url_path:
            args['url_path'] = parsed_args.url_path
        if parsed_args.http_method:
            args['http_method'] = parsed_args.http_method
        if parsed_args.name:
            args['name'] = parsed_args.name
        if parsed_args.disable:
            args['is_admin_state_up'] = False

        client = self.app.client_manager.network

        hm = client.find_health_monitor(name_or_id=parsed_args.health_monitor,
                                        ignore_missing=False)

        obj = client.update_health_monitor(
            health_monitor=hm.id,
            **args)

        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters=_formatters)

        return (display_columns, data)


class DeleteHealthMonitor(command.Command):
    _description = _('Delete a health monitor')

    def get_parser(self, prog_name):
        parser = super(DeleteHealthMonitor, self).get_parser(prog_name)

        parser.add_argument(
            'health_monitor',
            metavar='<health_monitor>',
            nargs='+',
            help=_('The ID of the health monitor to delete')
        )
        return parser

    def take_action(self, parsed_args):

        client = self.app.client_manager.network

        for hm in parsed_args.health_monitor:
            obj = client.find_health_monitor(name_or_id=hm,
                                             ignore_missing=False)
            client.delete_health_monitor(
                health_monitor=obj.id,
                ignore_missing=False)

        return



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
'''AS Instance v1 action implementations'''
import logging

from osc_lib import exceptions
from osc_lib import utils
from osc_lib.command import command

from otcextensions.i18n import _

LOG = logging.getLogger(__name__)


class ListAutoScalingInstance(command.Lister):
    _description = _('List AutoScaling Instances')
    columns = ('ID', 'Name', 'scaling_group_name',
               'scaling_configuration_id', 'scaling_configuration_name',
               'lifecycle_state', 'health_status', 'create_time')

    def get_parser(self, prog_name):
        parser = super(ListAutoScalingInstance, self).get_parser(prog_name)
        parser.add_argument(
            '--group',
            metavar='<group>',
            required=True,
            help=_('AS Group ID or Name for the instances query')
        )
        parser.add_argument(
            '--life-cycle-state',
            metavar='<life_cycle_state>',
            help=_('Life cycle state of the instances to query\n'
                   'Could be in [`INSERVICE`, `PENDING`, `REMOVING`]')
        )
        parser.add_argument(
            '--health-status',
            metavar='<health_status>',
            help=_('Health status of the instances to query\n'
                   'Could be in [`INITIALIZING`, `NORMAL`, `ERROR`]')
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            help=_('Limit the number of results displayed')
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.auto_scaling

        args = {}
        if parsed_args.limit:
            args['limit'] = parsed_args.limit
        if parsed_args.life_cycle_state:
            args['life_cycle_state'] = parsed_args.life_cycle_state
        if parsed_args.health_status:
            args['health_status'] = parsed_args.health_status

        group = client.find_group(parsed_args.group, ignore_missing=False)

        data = client.instances(group=group.id, **args)

        return (
            self.columns,
            (utils.get_item_properties(
                s,
                self.columns,
            ) for s in data)
        )


class RemoveAutoScalingInstance(command.Command):
    _description = _('Removes AutoScalinig Instances')

    def get_parser(self, prog_name):
        parser = super(RemoveAutoScalingInstance, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help=_('AS Instance ID to be deleted')
        )
        parser.add_argument(
            '--delete-instance',
            action='store_true',
            default=False,
            help=_('Specifies, whether instance should be completely deleted')
        )
        return parser

    def take_action(self, parsed_args):

        client = self.app.client_manager.auto_scaling
        client.remove_instance(
            instance=parsed_args.instance,
            delete_instance=parsed_args.delete_instance,
            ignore_missing=False
        )


class BatchActionAutoScalingInstance(command.Command):
    _description = _('Executes Action on AutoScalinig Instances')

    SUPPORTED_ACTIONS = ['ADD', 'REMOVE', 'PROTECT', 'UNPROTECT']

    def get_parser(self, prog_name):
        parser = super(BatchActionAutoScalingInstance, self).\
            get_parser(prog_name)
        parser.add_argument(
            '--group',
            metavar='<group>',
            required=True,
            help=_('AS Group ID or Name')
        )
        parser.add_argument(
            'action',
            metavar='<action>',
            help=_('Action to be performed\n'
                   'One of [`ADD`, `REMOVE`, `PROTECT`, `UNPROTECT`]')
        )
        parser.add_argument(
            'instance',
            metavar='<instance>',
            nargs='+',
            help=_('AS Instance ID to be deleted')
        )
        parser.add_argument(
            '--delete-instance',
            action='store_true',
            default=False,
            help=_('Specifies, whether instance should be completely deleted'
                   ' when action is ``REMOVE``')
        )
        return parser

    def take_action(self, parsed_args):

        action = parsed_args.action.upper()
        if action not in self.SUPPORTED_ACTIONS:
            msg = (_('Action %(action)s is not one of '
                     'the supported %(types)s') %
                   {'action': parsed_args.action,
                    'types': self.SUPPORTED_ACTIONS})
            raise exceptions.CommandError(msg)

        client = self.app.client_manager.auto_scaling
        client.batch_instance_action(
            group=parsed_args.group,
            instance=parsed_args.instance,
            action=action,
            delete_instance=parsed_args.delete_instance,
            ignore_missing=False
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.unit import base

from otcextensions.sdk.mrs.v1 import jobbinary

EXAMPLE = {
    "name": "my-job-binary",
    "url": "/simple/mapreduce/program",
    "is_protected": False,
    "is_public": False,
    "description": "this is the job binary template"
}


class TestJobBinary(base.TestCase):

    def test_basic(self):
        sot = jobbinary.Jobbinary()
        self.assertEqual('job_binary', sot.resource_key)
        self.assertEqual('binaries', sot.resources_key)
        path = '/job-binaries'
        self.assertEqual(path, sot.base_path)
        self.assertTrue(sot.allow_list)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_get)
        self.assertFalse(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_update)

    def test_make_it(self):
        sot = jobbinary.Jobbinary(**EXAMPLE)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['url'], sot.url)
        self.assertEqual(EXAMPLE['is_public'], sot.is_public)
        self.assertEqual(EXAMPLE['is_protected'], sot.is_protected)
        self.assertEqual(EXAMPLE['description'], sot.description)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack import resource
from openstack import utils


class ProtectionGroup(resource.Resource):
    """SDRS Protection Group Resource"""
    resource_key = 'server_group'
    resources_key = 'server_groups'
    base_path = '/server-groups'

    # capabilities
    allow_create = True
    allow_list = True
    allow_fetch = True
    allow_delete = True
    allow_commit = True

    _query_mapping = resource.QueryParameters(
        'availability_zone', 'limit', 'name', 'offset',
        'query_type', 'status')

    #: Properties
    #: Creation time
    created_at = resource.Body('created_at')
    #: Protection group description
    description = resource.Body('description')
    #: Number of DR drills in a protection group.
    disaster_recovery_drill_num = resource.Body('disaster_recovery_drill_num',
                                                type=int)
    #: Active-active domain ID
    domain_id = resource.Body('domain_id')
    #: Active-active domain name
    domain_name = resource.Body('domain_name')
    #: Disaster recovery type
    #: Default: migration
    dr_type = resource.Body('dr_type')
    #: Protection group health status
    health_status = resource.Body('health_status')
    #: Protection group ID
    id = resource.Body('id')
    #: Executed task job ID
    job_id = resource.Body('job_id')
    #: Protection group name
    name = resource.Body('name')
    #: Production site of protection group
    priority_station = resource.Body('priority_station')
    #: Protection group synchronization progress
    progress = resource.Body('progress', type=int)
    #: Number of protected instances in protection group
    protected_instance_num = resource.Body('protected_instance_num', type=int)
    # Specifies whether protection group is enabled
    protected_status = resource.Body('protected_status')
    #: Protection mode
    protection_type = resource.Body('protection_type')
    #: Reserved protection parameter
    replication_model = resource.Body('replication_model')
    #: Number of replication pairs in protection group
    replication_num = resource.Body('replication_num', type=int)
    #: Replication group data synchronization status
    replication_status = resource.Body('replication_status')
    #: Type of managed servers
    server_type = resource.Body('server_type')
    #: Production site AZ
    source_availability_zone = resource.Body('source_availability_zone')
    #: Production site VPC ID
    source_vpc_id = resource.Body('source_vpc_id')
    #: Protection group status
    status = resource.Body('status')
    #: Disaster recovery site AZ
    target_availability_zone = resource.Body('target_availability_zone')
    #: Disaster recovery site VPC ID
    target_vpc_id = resource.Body('target_vpc_id')
    #: DR drill VPC ID
    test_vpc_id = resource.Body('test_vpc_id')
    #: Update time
    updated_at = resource.Body('updated_at')

    def enable_protection_group(self, session, protection_group):
        """Method to enable protection for protection group

        :param session: The session to use for making this request.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param str protection_group: ID of protection group to
            be enabled
        """
        url = utils.urljoin(self.base_path, protection_group, '/action')
        body = {
            'start-server-group': {}
        }
        return session.post(url, json=body)

    def disable_protection_group(self, session, protection_group):
        """Method to disable protection for protection group

        :param session: The session to use for making this request.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param str protection_group: ID of protection group to
            be disabled
        """
        url = utils.urljoin(self.base_path, protection_group, '/action')
        body = {
            'stop-server-group': {}
        }
        return session.post(url, json=body)

    def perform_failover(self, session, protection_group):
        """Method to perform failover from production site to
        DR site

        :param session: The session to use for making this request.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param str protection_group: ID of protection group to
            perform failover
        """
        url = utils.urljoin(self.base_path, protection_group, '/action')
        body = {
            'failover-server-group': {}
        }
        return session.post(url, json=body)

    def perform_planned_failover(self, session, protection_group,
                                 priority_station):
        """Method to perform planned failover from production site
        to DR site

        :param session: The session to use for making this request.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param str protection_group: ID of protection group to
            perform planned failover
        :param str priority_station: direction of planned failover
            Values: source of target
        """
        url = utils.urljoin(self.base_path, protection_group, '/action')
        body = {
            "reverse-server-group": {
                "priority_station": priority_station
            }
        }
        return session.post(url, json=body)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
'''Anti DDoS Floating IP v1 action implementations'''
import logging

from osc_lib import utils
from osc_lib.command import command

from otcextensions.i18n import _
from otcextensions.common import sdk_utils

LOG = logging.getLogger(__name__)


STATUS_VALUES = ['normal', 'configging', 'notConfig',
                 'packetcleaning', 'packetdropping']


_formatters = {
}


def _get_columns(item):
    column_map = {
    }
    return sdk_utils.get_osc_show_columns_for_sdk_resource(item, column_map)


class ListFloatingIP(command.Lister):
    _description = _('List Anti DDoS FloatingIP statuses')
    columns = (
        'floating_ip_address', 'floating_ip_id', 'network_type', 'status'
    )

    def get_parser(self, prog_name):
        parser = super(ListFloatingIP, self).get_parser(prog_name)

        parser.add_argument(
            '--ip',
            metavar='<IP>',
            help=_('Floating IP prefix.')
        )
        parser.add_argument(
            '--status',
            metavar='{' + ','.join(STATUS_VALUES) + '}',
            type=lambda s: s.lower(),
            choices=STATUS_VALUES,
            help=_('Policy status value.')
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.anti_ddos

        query = {}

        if parsed_args.ip:
            query['ip'] = parsed_args.ip
        if parsed_args.status:
            query['status'] = parsed_args.status

        data = client.floating_ips(**query)

        table = (self.columns,
                 (utils.get_item_properties(
                     s, self.columns
                 ) for s in data))
        return table


class ShowFloatingIP(command.ShowOne):
    _description = _('Show the policies of a single floating IP')

    def get_parser(self, prog_name):
        parser = super(ShowFloatingIP, self).get_parser(prog_name)

        parser.add_argument(
            'floating_ip_id',
            metavar='<ID>',
            help=_('UUID of the floating IP.')
        )

        return parser

    def take_action(self, parsed_args):

        client = self.app.client_manager.anti_ddos

        obj = client.get_floating_ip_policies(
            parsed_args.floating_ip_id,
        )

        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)

        return (display_columns, data)


class SetFloatingIP(command.ShowOne):
    _description = _('Update a single floating ip instance policies')

    def get_parser(self, prog_name):
        parser = super(SetFloatingIP, self).get_parser(prog_name)
        parser.add_argument(
            'floating_ip_id',
            metavar='<ID>',
            help=_('UUID of the floating IP.')
        )
        parser.add_argument(
            '--traffic_pos_id',
            type=int,
            metavar='<[1..9]>',
            choices=range(1, 10),
            required=True,
            help=_('Position ID of traffic. The value ranges from 1 to 9.')
        )
        parser.add_argument(
            '--http_request_pos_id',
            type=int,
            metavar='<[1..15]>',
            choices=range(1, 16),
            required=True,
            help=_('Position ID of number of HTTP requests. '
                   'The value ranges from 1 to 15.')
        )
        parser.add_argument(
            '--cleaning_access_pos_id',
            type=int,
            metavar='<[1..8]>',
            choices=range(1, 9),
            required=True,
            help=_('Position ID of number of access limit during cleaning. '
                   'The value ranges from 1 to 8.')
        )
        parser.add_argument(
            '--app_type_id',
            metavar='<[0..1]>',
            type=int,
            required=True,
            help=_('Application type ID.')
        )
        l7_group = parser.add_mutually_exclusive_group()
        l7_group.add_argument(
            '--enable_l7',
            action='store_true',
            default=True,
            help=_('Enable L7 defence (default).')
        )
        l7_group.add_argument(
            '--disable_l7',
            action='store_true',
            help=_('Disable L7 defence.')
        )

        return parser

    def take_action(self, parsed_args):

        client = self.app.client_manager.anti_ddos

        attrs = {}

        if parsed_args.enable_l7:
            attrs['is_enable_l7'] = True
        if parsed_args.disable_l7:
            attrs['is_enable_l7'] = False
        if parsed_args.traffic_pos_id:
            attrs['traffic_pos_id'] = parsed_args.traffic_pos_id
        if parsed_args.http_request_pos_id:
            attrs['http_request_pos_id'] = parsed_args.http_request_pos_id
        if parsed_args.cleaning_access_pos_id:
            attrs['cleaning_access_pos_id'] = \
                parsed_args.cleaning_access_pos_id
        if parsed_args.app_type_id:
            attrs['app_type_id'] = parsed_args.app_type_id

        obj = client.update_floating_ip_policies(
            floating_ip=parsed_args.floating_ip_id, **attrs)

        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters=_formatters)

        return (display_columns, data)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import uuid

from openstack import _log

from otcextensions.sdk.vpc.v1 import vpc
from otcextensions.tests.functional import base

_logger = _log.setup_logging('openstack')


class TestService(base.BaseFunctionalTest):
    ID = None
    uuid = uuid.uuid4().hex[:8]

    def setUp(self):
        super(TestService, self).setUp()

        attrs = {
            'name': "test-vpc-" + self.uuid,
            'cidr': '192.168.0.0/24'
        }

        self.NAME = "test-vpc-" + self.uuid
        self.UPDATE_NAME = "test-vpc-upd-" + self.uuid

        self.vpc = self.conn.vpc.create_vpc(**attrs)
        assert isinstance(self.vpc, vpc.Vpc)
        self.assertEqual(self.NAME, self.vpc.name)
        self.ID = self.vpc.id
        self.addCleanup(self.conn.vpc.delete_vpc, self.vpc)

    def test_find_vpc(self):
        found = self.conn.vpc.find_vpc(self.NAME)
        self.assertEqual(found.id, self.ID)

    def test_get_vpc(self):
        found = self.conn.vpc.get_vpc(self.ID)
        self.assertEqual(found.name, self.NAME)
        self.assertEqual(found.id, self.ID)

    def test_list_vpcs(self):
        vpcs = [o.name for o in self.conn.vpc.vpcs()]
        self.assertIn(self.NAME, vpcs)

    def test_update_vpc(self):
        new_attrs = {
            'name': self.UPDATE_NAME,
            'cidr': '192.168.0.0/16'
        }
        updated = self.conn.vpc.update_vpc(self.ID, **new_attrs)
        self.assertEqual(updated.name, new_attrs['name'])
        self.assertEqual(updated.cidr, new_attrs['cidr'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack import resource


class StorageType(resource.Resource):

    base_path = '/storage-type/%(datastore_name)s'
    resources_key = 'storage_type'

    # capabilities
    allow_list = True
    _query_mapping = resource.QueryParameters(
        'version_name')

    #: DB engine name
    #: *Type: str*
    datastore_name = resource.URI('datastore_name')
    #: Indicates the storage type
    #: *Type: str*
    name = resource.Body('name')
    #: Indicates the performance specifications
    #: *Type: list*
    compute_group_type = resource.Body('support_compute_group_type', type=list)
    #: Availability Zone Status
    #: *Type: dict*
    az_status = resource.Body('az_status', type=dict)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import binascii
import hashlib

from openstack import resource

from otcextensions.sdk.kms.v1 import _base


def calculate_plain_message(plain_text):
    hash = hashlib.sha256()
    hex_data = binascii.unhexlify(plain_text)
    # hex_data = str(plain_text).decode(encoding='hex')
    hash.update(bytearray(hex_data))
    digest = hash.hexdigest()
    return (plain_text + digest, len(hex_data))


class DataKey(_base.Resource):

    create_path = '/kms/create-datakey'

    allow_create = True
    allow_update = True
    allow_fetch = False

    # Properties
    #: Secret key (CMK) ID
    #: *Type:str*
    key_id = resource.Body('key_id')
    #: Encryption context
    #: Dict (as string) describing resource context information
    #: *Type:str*
    encryption_context = resource.Body('encryption_context', type=dict)
    #: Datakey length
    #: *Type:int*
    datakey_length = resource.Body('datakey_length')
    #: Datakey plain length
    #: *Type:int*
    datakey_plain_length = resource.Body('datakey_plain_length')
    #: Datakey plain length
    #: *Type:int*
    datakey_cipher_length = resource.Body('datakey_cipher_length')
    #: Sequence
    #: *Type:str*
    sequence = resource.Body('sequence')
    #: Plain text of the data key
    #: *Type:str*
    plain_text = resource.Body('plain_text')
    #: Cipher text of the data key and encrypted_value
    #: *Type:str*
    cipher_text = resource.Body('cipher_text')

    def create_wo_plain(self, session, prepend_key=True, requires_id=True):
        return super(DataKey, self).create(
            session, prepend_key=prepend_key,
            uri='/kms/create-datakey-without-plaintext',
            requires_id=requires_id)

    def encrypt(self, session):
        """Encrypt DEK

        Encrypt DEK with CMK. Requires plain_text to be set with the DEK's
            plain value. Populates encrypted_cipher_text attribute
            with the generated cipher in hex (2 chars per byte)
        """
        session = self._get_session(session)

        plain_text = self.plain_text
        if plain_text is None:
            raise ValueError('plain_text should be provided')
        (msg, size) = calculate_plain_message(plain_text)
        # params.update({'plain_text': msg})
        body = {
            'key_id': self.key_id,
            'plain_text': msg,
            'datakey_plain_length': size,
        }
        if self.sequence:
            body['sequence'] = self.sequence
        if self.encryption_context:
            body['encryption_context'] = self.encryption_context

        url = '/kms/encrypt-datakey'
        response = session.post(
            url,
            json=body)

        self._translate_response(response)

        self._update(
            datakey_cipher_length=self.datakey_length
        )

        return self

    def decrypt(self, session):
        """Decrypt Key

        Decrypts `cipher_text` into `plain_text` with the given CMK
        """
        session = self._get_session(session)
        body = {
            'key_id': self.key_id,
            'cipher_text': self.cipher_text,
            'datakey_cipher_length': self.datakey_cipher_length,
        }
        if self.sequence:
            body['sequence'] = self.sequence
        if self.encryption_context:
            body['encryption_context'] = self.encryption_context

        url = '/kms/decrypt-datakey'
        response = session.post(
            url,
            json=body)

        self._translate_response(response)
        data = response.json()
        self._update(
            plain_text=data['data_key']
        )

        return self



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from otcextensions.tests.functional.sdk.vlb import TestVlb


class TestListener(TestVlb):

    def setUp(self):
        super(TestListener, self).setUp()
        self.create_network()
        self.create_load_balancer()
        self.create_listener()

    def test_01_list_listeners(self):
        lst = list(self.client.listeners())
        self.assertIsNotNone(lst)

    def test_02_get_listener(self):
        lst = self.client.get_listener(TestVlb.listener)
        self.assertIsNotNone(lst)

    def test_03_find_listener(self):
        lst = self.client.find_listener(TestVlb.listener.name)
        self.assertIsNotNone(lst)

    def test_05_update_listener(self):
        new_description = 'changed'
        lst = self.client.update_listener(
            TestVlb.listener,
            description=new_description
        )
        self.assertEqual(lst['description'], new_description)

        # cleanup
        self.client.delete_listener(
            TestVlb.listener
        )
        self.client.delete_load_balancer(
            TestVlb.load_balancer
        )
        self.net_client.delete_ip(
            TestVlb.load_balancer.floating_ips[0]['publicip_id']
        )
        self.addCleanup(self.destroy_network, TestVlb.network)
